#! /usr/bin/env python3

from action_graph.action import Action, ActionStatus, State
from action_graph.agent import Agent

name = 'action_graph'
__version__ = '1.2.5'
__all__ = [
    'State',
    'Action',
    'ActionStatus',
    'ActionFailedException',
    'ActionAbortedException',
    'ActionTimedOutException',
    'Planner',
    'PlanningFailedException',
    'Agent',
    'name'
    '__version__',
]

# logging setup
import logging

console = logging.StreamHandler()
console.setLevel(logging.INFO)
#
formatter = logging.Formatter('>>>%(levelname)s > %(message)s')
console.setFormatter(formatter)
#
logging.getLogger('').addHandler(console)
logger = logging.getLogger(__name__)



#! /usr/bin/env python3

from action_graph.action import Action
from action_graph.agent import Agent


class Action1(Action):
    effects = {"FIRST": True}
    preconditions = {}


class Action2A(Action):
    effects = {"SECOND": True}
    preconditions = {"FIRST": True}


class Action2B(Action):
    effects = {"SECOND": True}
    preconditions = {}
    cost = 1.5


class Action3(Action):
    effects = {"THIRD": True}
    preconditions = {"FIRST": True, "SECOND": True}


def test():
    world_state = {"FIRST": False, "SECOND": False, "THIRD": False}
    goal_state = {"THIRD": True}

    ai = Agent()

    actions = [a(ai) for a in Action.__subclasses__()]
    ai.load_actions(actions)

    print("Initial State:", world_state)
    ai.update_state(world_state)

    print("Goal State:   ", goal_state)
    plan = ai.get_plan(goal_state)

    expected_actions = ["Action1", "Action2B", "Action3"]
    expected_outcome = [{'FIRST': True}, {'SECOND': True}, {'THIRD': True}]

    for ax, eax, eoc in zip(plan, expected_actions, expected_outcome):
        assert ax.__class__.__name__ == eax, f'Incorrect Action!'
        assert ax.effects == eoc, f'Incorrect Action Outcome!'




#! /usr/bin/env python3

from action_graph.action import Action, State, ActionStatus
from action_graph.agent import Agent


class FibonacciIncrement(Action):
    effects = {"counter": ...}

    def apply_effects(self, outcome: State, state: State):
        # store the sum in the state
        state["fibonacci_sum"] = state["fibonacci_sum"] + state["counter"] + 1
        # now apply effects
        # dont apply the final effect(base class implementation);
        # instead increment by 1
        state["counter"] = state["counter"] + 1


def test():
    world_state = {"fibonacci_sum": 0, "counter": 0}
    goal_state = {"counter": 10}

    ai = Agent()
    actions = [a(ai) for a in Action.__subclasses__()]
    ai.load_actions(actions)
    ai.update_state(world_state)
    for plan in ai.achieve_goal_interactive(goal_state):
        pass

    assert ai.state["fibonacci_sum"] == 55, f'Looping feature broken!'



#! /usr/bin/env python3

import logging
from time import sleep, time
from typing import List

from action_graph.action import (Action, ActionStatus, State,
                                 ActionTimedOutException, ActionAbortedException, ActionFailedException)
from action_graph.planner import Planner, PlanningFailedException


class Agent:
    """Autonomous agent to monitor system state, keep track of feasible actions, generate plans and achive desired goals"""

    __planner: Planner = Planner([])
    __abort: bool = False

    def __init__(self, agent_name=None) -> None:
        if not agent_name:
            agent_name = self.__class__.__name__
        self.name = agent_name
        #
        self.state: State = {}

    def load_actions(self, actions: List[Action]):
        """
        Load actions list; refresh in case of any changes to the actions data.

        :param actions:List[Action]: List of actions.
        """

        self.__planner.update_actions(actions)

    def update_state(self, state: State):
        """
        Updates system state with the incoming state.        

        :param state:State: New state
        """

        self.state.update(state)

    def abort(self):
        """
        Abort execution.
        """

        self.__abort = True

    def reset(self):
        """
        Resets the abort status back to False in the event abort was triggered.
        """

        self.__abort = False

    def is_goal_met(self, goal: State) -> bool:
        """
        Compares the desired goal state against the system state to check if it is already met.

        :param goal:State: Desired goal state
        :return:bool: True if goal state and current system state is same; False otherwise
        """

        if not goal:
            return True

        for k, v in goal.items():
            if k not in self.state or self.state[k] != v:
                return False

        return True

    def get_plan(self, goal: State, start_state: State = None, actions: List[Action] = None) -> List[Action]:
        """
        Generate an action plan for the specified goal state.
        If no start_state is provided, the current state of the system is used. 

        :param goal:State: Specify the goal state.
        :param start_state:State=None: Specify a start state that is not the current state.
        :param actions:List[Action]=None: List of actions that the planner can use; 
                                          If this is not specified, the planner will use the 
                                          previously loaded actions to plan.
        :return:List[Action]: The plan - dictionary of actions and their expected outcomes. 
        """

        if not start_state:
            start_state = self.state

        if actions:
            self.__planner.update_actions(actions)

        try:
            return self.__planner.generate_plan(goal, start_state)
            #
        except PlanningFailedException as pfx:
            logging.error(f"PLANNING FAILED! {pfx}")
            return []

    def execute_plan(self, plan: List[Action]):
        """
        Execute a previously generated plan.

        :param plan:List[Action]: Dictionary of actions and their expected outcomes (State)
        """

        for action in plan:
            try:
                self.execute_action(action)

            except Exception as _ex:
                logging.error(f"{_ex}")
                raise

        logging.info(f"EXECUTION SUCCEDED!")

    def achieve_goal(self, goal: State, verbose: bool = False):
        """
        Creates a plan to satisfy the goal state; executes it action-by-action; re-evaluates the plan at each step;

        :param goal:State: Desired goal state
        :param verbose:bool: Print plan to console at each step, if True
        """

        __actions_to_avoid: List[str] = []

        # state might have changed since the last step was executed
        while not self.is_goal_met(goal):

            try:
                # (re)generate the plan
                plan: List[Action] = self.__planner.generate_plan(goal, self.state, __actions_to_avoid)
                if verbose:
                    self.print_plan_to_console(plan)
                # execute one plan step at a time
                __action = plan[0]
                self.execute_action(__action)

            except ActionFailedException as ex_fail:
                logging.error(f"{ex_fail} / ATTEMPTING ALTERNATIVE PLAN")
                __action_name = str(__action)
                if __action_name not in __actions_to_avoid:
                    __actions_to_avoid.append(__action_name)
                continue

            except Exception as _ex:
                logging.error(f"{_ex}")
                raise

        logging.info(f"EXECUTION SUCCEDED!")

    def achieve_goal_interactive(self, goal: State):
        """
        Creates a plan to satisfy the goal state; executes it action-by-action; re-evaluates the plan at each step;

        :param goal:State: Desired goal state
        """

        __actions_to_avoid: List[str] = []

        # state might have changed since the last step was executed
        while not self.is_goal_met(goal):

            try:
                # (re)generate the plan
                plan: List[Action] = self.__planner.generate_plan(goal, self.state, __actions_to_avoid)
                yield plan
                # execute one plan step at a time
                self.execute_action(plan[0])

            except ActionFailedException as ex_fail:
                logging.error(f"{ex_fail} / ATTEMPTING ALTERNATIVE PLAN")
                __action_name = str(plan[0])
                if __action_name not in __actions_to_avoid:
                    __actions_to_avoid.append(__action_name)
                continue

            except Exception as _ex:
                logging.error(f"{_ex}")
                raise

        logging.info(f"EXECUTION SUCCEDED!")

    def print_plan_to_console(self, plan: List[Action]):
        if plan:
            plan_str = '\nPLAN:\n'
            for ix, action in enumerate(plan):
                plan_str += "----- "*(ix+1) + str(action) + ' -- ' + str(action.effects) + '\n'
            print(plan_str)

    def execute_action(self, action: Action):
        # Check for abort status
        if self.__abort:
            # logging.error(f'ACTION: {action} : EXECUTION ABORTED BEFORE START !!')
            action.on_aborted(action.effects)
            raise ActionAbortedException(f'ACTION: {action} FAILED. ABORTED STATE IS ACTIVE!!')

        # Check runtime precondition
        if not action.check_runtime_precondition(action.effects):
            raise ActionFailedException(f'ACTION: {action} RUNTIME PRECONDITION CHECK FAILED!!.')

        # Execute the plan step
        action._execute(action.effects)
        # action.execute is an async process inside _execute,

        if action.allow_async:
            # if this is an async action; just apply the effects and return
            action.apply_effects(action.effects, self.state)
            return

        # monitor the status; wait until execution is complete
        time0 = time()
        while action.is_running():
            if self.__abort:
                # if an abort was signalled
                logging.critical(f'ACTION: {action} : EXECUTION ABORTED!!')
                action.status = ActionStatus.ABORTED
                break
            if time()-time0 > action.timeout:
                # Action timeout exceeded
                raise ActionTimedOutException(f'ACTION: {action} : TIMED OUT!!')
            if not action.status == ActionStatus.RUNNING:
                # thread is alive but the status has changed
                break  # so move on
            sleep(0.05)  # throttle loop
            # logging.debug(f'Action: {str(action)} is running...')

        # Execution completed but with RUNNING Status
        if action.status == ActionStatus.RUNNING:
            # the user forgot to set the status; or something bad happened;
            # let's treat this as FAILURE!
            action.status = ActionStatus.FAILURE
            logging.warning(f'ACTION: {action} STATUS UNKNOWN; ASSUMING FAILURE!!')

        # Execution completed with NEUTRAL Status
        if action.status == ActionStatus.NEUTRAL:
            # this is when action completes with no change to the state
            # logging.debug(f'ACTION: {action} Action completed with neutral state.')
            action.on_neutral(action.effects)  # ignore effects

        # Execution completed with SUCCESS Status
        if action.status == ActionStatus.SUCCESS:
            # Action executed without errors;
            action.apply_effects(action.effects, self.state)
            # logging.debug(f'ACTION: {action} Action succeded.')
            action.on_success(action.effects)

        # Execution failed
        if action.status == ActionStatus.FAILURE:
            action.on_failure(action.effects)
            # Stop execution as action failed
            raise ActionFailedException(f'ACTION: {action} FAILED!')

        # Execution aborted
        if action.status == ActionStatus.ABORTED:
            action.on_aborted(action.effects)
            # Stop execution as action aborted
            raise ActionAbortedException(f'ACTION: {action} ABORTED!')

        # Any clean up needed after execution e.g. updating system states
        action.on_exit(action.effects)



#! /usr/bin/env python3

import sys
from collections import defaultdict
from typing import Any, Dict, List, Tuple

from action_graph.action import Action, State, ImpossibleAction


class PlanningFailedException(Exception):
    pass


class Planner():
    """Search and determine a plan (sequence of actions) that satifies a desired goal state"""

    def __init__(self, actions: List[Action]) -> None:
        self.update_actions(actions)

    def update_actions(self, actions: List[Action]):
        """
        The refereshes/reloads the list of Actions available to the Planner.

        :param actions:List[Action]: List of actions (instances of Action class)
        """

        self._action_lookup: defaultdict = self.__create_action_lookup(actions)

    def generate_plan(self, target_state: State, start_state: State, avoid_actions: List[Action] = None) -> List[Action]:
        """
        Find and return an optimal sequence of actions (the plan) that will 
        lead from the start state to the target state.

        :param target_state:State: Desired goal (target) state
        :param start_state:State: Current/start state of the system
        :return:List[Action]: List of actions updated with their expected outcomes (effects)
        """

        if len(target_state.items()) > 1:
            raise PlanningFailedException('target_state should have a single state variable')
        tk, tv = list(target_state.items())[0]
        # in case target state value is a reference to another state variable
        tv = self.__parse_references(tv, start_state)
        # check if the target state is already satisfied
        if (tk, tv) in list(start_state.items()):
            return []   # goal already met, move on
        #
        # find action(s) that satisfy the state current effect-item
        probable_actions: List[Action] = self._action_lookup[(tk, tv)]
        if not probable_actions:  # if no actions are found, try with templated actions
            probable_actions = self._action_lookup[(tk, Ellipsis)]
        if avoid_actions:  # actions we do not want to consider for planning
            probable_actions = [a for a in probable_actions if str(a) not in avoid_actions]
        if not probable_actions:
            return [ImpossibleAction(effects={tk: tv})]

        chosen_path: List[Action] = []
        for p_action in probable_actions:  # explore each available action...
            action = p_action.__copy__()
            if action.effects[tk] is Ellipsis:
                action.effects[tk] = tv  # apply variable effects
            #
            action_path: List[Action] = []
            for pk, pv in action.preconditions.items():  # for each pre-condition ...
                try:  # choose the shortest feasible path
                    pv = self.__parse_references(pv, action.effects)
                    action_path += self.generate_plan({pk: pv}, start_state, avoid_actions)  # merge the actions
                except RecursionError:  # watch out for cyclic references
                    raise PlanningFailedException(f'Found cyclic references!')
            # include the current action;  remove duplicates; keep the order intact
            action_path = self.__make_unique(action_path + [action])
            #
            if not chosen_path:  # if no other path is available...
                chosen_path = action_path  # use the current path
                continue
            # if alternative paths exist, use the one with the lowest cost
            if sum(a.cost for a in action_path) < sum(a.cost for a in chosen_path):
                chosen_path = action_path

        # check if path is feasible; path cost should be < infinite cost
        impossible_actions = [a for a in chosen_path if a.cost >= sys.float_info.max]
        for ia in impossible_actions:
            raise PlanningFailedException(f'No action available to satisfy: {ia.effects}')
        return chosen_path

    def __create_action_lookup(self, actions: List[Action]) -> Dict[Tuple[Any, Any], List[Action]]:
        action_lookup: Dict[Tuple[Any, Any], List[Action]] = defaultdict(list)
        for action in actions:
            for k, v in action.effects.items():
                action_lookup[(k, v)].append(action)
        return action_lookup

    def __parse_references(self, ref: Any, state: State) -> Any:
        if isinstance(ref, str) and ref[0] == '$':
            if ref[1:] in state:
                return state[ref[1:]]
        return ref

    def __make_unique(self, path):
        unique = set()
        return [x for x in path if x not in unique and not unique.add(x)]
